'From Squeak3.0 of 4 February 2001 [latest update: #3552] on 1 July 2003 at 10:32:26 am'!
Fraction subclass: #Decimal
	instanceVariableNames: 'decimalPlaces '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Kernel-Numbers'!

!Decimal commentStamp: 'gzi 7/1/2003 10:32' prior: 0!
This class serves as an implementation of decimal numbers as they are used by SQL databases.
The main advantage of using instances of this class is that there are no rounding errors.

Usage:

| aDecimal2 aDecimal4 aDecimal10 some double large added |
"creating instances"
aDecimal2 := Decimal fromString:'12.34' .
aDecimal10 := '1234567890.234567890' asDecimal .
aDecimal4 := Decimal numerator:2345 denominator:100 decimalPlaces:4 .

"printout"
Transcript cr; show:'aDecimal2 :'; show: (aDecimal2 printString).
Transcript cr; show:'aDecimal10:'.  aDecimal10 printOn: Transcript.
Transcript cr; show:'aDecimal4 :'; show: (aDecimal4 printPaddedTo:23).
Transcript cr; show:'aDecimal10:'; show: (aDecimal10 displayStringFormatted).
Transcript cr.

"arithmetics"
double := aDecimal2 * 2.
some := aDecimal4 / 3.1415926.
large := aDecimal10 * aDecimal4.
added := aDecimal2 - aDecimal4.

"check your Transcripts font-stype here: should be fixed"
double decimalPlaces:4.
Transcript cr; show: (double printPaddedTo:18). 
some decimalPlaces:4.
Transcript cr; show: (some printPaddedTo:18).
large decimalPlaces:4.
Transcript cr; show: (large printPaddedTo:18).
added decimalPlaces:4.
Transcript cr; show: (added printPaddedTo:18).


"Check the difference !!"
Transcript cr; show:'Float  :'; show:(
(                       ( 0.0000001   +
                         1000000)     -
                         1000000
                         ) printString).
Transcript cr; show:'Decimal:'; show: (
(((Decimal fromString:'0.0000001') + 
(Decimal fromString:'1000000') )   -
(Decimal fromString:'1000000')
                         ) printString).
!


!Number methodsFor: 'as yet unclassified' stamp: 'gzi 7/1/2003 08:01'!
asDecimal: decimalDigits
		"Create a new instance of Decimal and initialize it with the value of the receivers"
	| shift |
	(self floor = self) ifTrue:[ "no decimal fraction"
		^Decimal numerator: self numerator
		  denominator: self denominator
		  decimalPlaces: decimalDigits.
	] ifFalse:[
		shift := (10 raisedTo:decimalDigits).
		^Decimal numerator: (((self numerator) * shift) truncated)
		denominator: shift
		decimalPlaces: decimalDigits.
	].
		
! !

!Number methodsFor: 'as yet unclassified' stamp: 'gzi 7/1/2003 08:01'!
denominator

	"Answer a number which represents the
	 denominator of the receiver."

	 ^1 ! !

!Number methodsFor: 'as yet unclassified' stamp: 'gzi 7/1/2003 08:01'!
numerator

	"Answer a type of Number which represents the
	 numerator of the receiver."

	 ^self ! !


!Fraction methodsFor: 'as yet unclassified' stamp: 'gzi 7/1/2003 08:01'!
asDecimal: decimalDigits
		"Create a new instance of Decimal and initialize it with the value of the receivers"
	
		^Decimal numerator: self numerator
		  denominator: self denominator
		  decimalPlaces: decimalDigits.
	
		
! !


!Decimal methodsFor: 'as yet unclassified' stamp: 'gzi 7/1/2003 08:01'!
* aNumber
		"Answer the result of multiplying
		 the receiver by aNumber."
	^super * aNumber
			asDecimal: decimalPlaces.! !

!Decimal methodsFor: 'as yet unclassified' stamp: 'gzi 7/1/2003 08:01'!
+ aNumber
		"Answer sum of the receiver and aNumber."
	^super + aNumber
			asDecimal: decimalPlaces.! !

!Decimal methodsFor: 'as yet unclassified' stamp: 'gzi 7/1/2003 08:01'!
- aNumber
		"Answer the difference between
		 the receiver and aNumber."
	^super - aNumber
			asDecimal: decimalPlaces.! !

!Decimal methodsFor: 'as yet unclassified' stamp: 'gzi 7/1/2003 08:01'!
/ aNumber
		"Answer the result of dividing
		 the receiver by aNumber."
	^super / aNumber
			asDecimal: decimalPlaces.! !

!Decimal methodsFor: 'as yet unclassified' stamp: 'gzi 7/1/2003 08:01'!
= aNumber
		"Answer true if the receiver is equal
		 to aNumber, else answer false."
	^super = aNumber ! !

!Decimal methodsFor: 'as yet unclassified' stamp: 'gzi 7/1/2003 08:01'!
decimalPlaces
	"Answer the number of decimal digits behind the decimal sign
	This is for formatting porposes only."

	^decimalPlaces! !

!Decimal methodsFor: 'as yet unclassified' stamp: 'gzi 7/1/2003 08:01'!
decimalPlaces: aPositiveInteger
	"Set the number of decimal digits behind the decimal sign"
	decimalPlaces:= aPositiveInteger.! !

!Decimal methodsFor: 'as yet unclassified' stamp: 'gzi 7/1/2003 08:01'!
decimalPlacesAdjust
		"return an instance of Decimal that is the receiver adjusted to
		 aNumber decimal places"
	|decDenominator exact trunc fract|
	decDenominator:= 10 raisedToInteger: decimalPlaces.
	exact:= decDenominator / denominator * numerator.
	trunc:= exact truncated.
	fract:= (exact - trunc) abs.
	numerator:= trunc abs.
	denominator:= decDenominator.
	(fract >= (1 / 2))
		ifTrue: [numerator:= numerator + 1].
	(exact negative)
		ifTrue: [numerator:= 0 - numerator].
	^self.! !

!Decimal methodsFor: 'as yet unclassified' stamp: 'gzi 7/1/2003 08:01'!
displayStringFormatted
	"returns an external representation of the receiver"

	| s ws rs rest pos |

	s := self printString copyReplaceAll: '.' with: ','.

	"insert grouping commas"
	ws := WriteStream on: ( String new ).
	rs := ReadStream on: ( ( ReadStream on: ( s upTo: $, ) ) contents reverse ).
	( ( s upTo: $, ) reject: [ :c | c = $-. ] ) size - 1 // 3 timesRepeat: [
		ws
			nextPutAll: ( rs next: 3 );
			nextPut: $..
	].
	[ rs atEnd. ] whileFalse: [ ws nextPut: ( rs next ). ].
	rs := ReadStream on: ( ws contents ).
	( pos := s indexOf: $, ) = 0
		ifTrue: [ rest := ''. ]
		ifFalse: [ rest := s copyFrom: pos to: ( s size ). ].
	s := rs contents reverse , rest.

	( s at: s size ) = $,
		ifTrue: [ s := s copyFrom: 1 to: ( s size - 1 ). ].

	^ s! !

!Decimal methodsFor: 'as yet unclassified' stamp: 'gzi 7/1/2003 08:01'!
negated
		"Answer an instance of class Decimal
		 which is the negative of the receiver."
	^super negated
		asDecimal: decimalPlaces.! !

!Decimal methodsFor: 'as yet unclassified' stamp: 'gzi 7/1/2003 08:01'!
printOn: aStream
		"writes the receiver to aStream"

	self deepCopy printOnAdjusted: aStream! !

!Decimal methodsFor: 'as yet unclassified' stamp: 'gzi 7/1/2003 08:01'!
printOnAdjusted: aStream
		"writes the receiver adjusted to aStream"
	|tempStream numDigits|

	self decimalPlacesAdjust.

	tempStream:= ReadWriteStream on: (String new: 15).
	numerator abs printOn: tempStream.
	numDigits:= tempStream position.
	tempStream reset.

	"sign"
	(self < 0) ifTrue: [
		aStream nextPut: $-
	].

	(decimalPlaces >= numDigits) ifTrue: [
			"zeroes in front of digits"
			aStream nextPutAll: '0.'.
			(decimalPlaces - numDigits) timesRepeat: [
				aStream nextPut: $0
			]
	] ifFalse: [
			"digits in front of decimal point"
			aStream nextPutAll: (tempStream next:  numDigits - decimalPlaces);
			nextPut: $.
	].
	aStream nextPutAll: (tempStream next:  numDigits - tempStream position).! !

!Decimal methodsFor: 'as yet unclassified' stamp: 'gzi 7/1/2003 08:01'!
printPaddedTo: anInteger
		"Answer a string containing the ASCII
		 representation of the receiver padded
		 on the left with blanks to be at least
		 anInteger characters."
	| aStream padding digits|
	aStream := WriteStream on: (String new: 10).
	self printOn: aStream.
	digits := aStream contents.
	padding := anInteger - digits size.
	padding > 0
		ifTrue: [
			padding := ((String new: padding)
				atAllPut: $ ;
				yourself)]
		ifFalse:[padding := ''].
	^padding, digits! !


!Decimal class methodsFor: 'as yet unclassified' stamp: 'gzi 7/1/2003 09:04'!
fromString: aString 
		"creates an instance of the receiver class that is initialized with the contents of the
string. The string my contain a leading minus sign, a decimal point and grouping commas.
Test:

self halt.	
Decimal fromString:'3.14159'

"
	|
	point
	integer
	fraction
	start
	end
	dPlaces
	last
	nst             "normalized String"
	result
	decimalPlaces |

	"remove all grouping commas or points except the last"
	last := true.
	nst := ''.
	decimalPlaces := 0.
	aString reverseDo:[:c|
		((c = $.) or:[c = $,]) ifTrue:[
			last ifTrue:[
				"keep the last , forget the rest"
				nst := '.',nst.
				last := false.
			].
		] ifFalse:[
			nst := (c asString), nst.
		].
		last ifTrue:[ "just behind the decimal point"
			decimalPlaces := decimalPlaces + 1.
		].
	].

	start:= 1.

	end:= nst size.
	(end = 0) ifTrue:[ "empty string"
		integer := 0.
		^self numerator: 0 denominator: 1 decimalPlaces: decimalPlaces
	].

	((nst at: 1) = $-) ifTrue: [ "leading minus sign"
		start:= 2
	].


	point:= nst indexOf: $.  ifAbsent: [
		  integer:= (nst copyFrom: start to: end) asInteger.
		  ((nst at: 1) = $-) ifTrue: [
				integer:= integer negated
		  ].
		 ^self numerator: integer denominator: 1 decimalPlaces: 0
	].
	(point = start)  ifTrue: [
		integer:= 0
	] ifFalse: [
		integer:= (nst copyFrom: start  to: (point - 1)) asInteger
	].
	start:= point + 1.
	dPlaces:= end - point.
	(dPlaces = 0) ifTrue: [
		fraction:= 0
	] ifFalse: [
		fraction:= (nst copyFrom: start to: end) asInteger
	].
	result:= (Decimal numerator: fraction
		  denominator: (10 raisedToInteger: dPlaces) decimalPlaces:decimalPlaces)
		  + integer .
	((nst at: 1) = $-) ifTrue: [
		result:= result negated 
	].
	^result
! !

!Decimal class methodsFor: 'as yet unclassified' stamp: 'gzi 7/1/2003 08:01'!
numerator: n denominator: d decimalPlaces: p
		"Answer an instance of class Decimal and
		 initialize both numerator and denominatior
		 instance variables to n and d respectively.
		 Initialize decimalPlaces to p."
	^(self numerator: n denominator: d) decimalPlaces: p.! !


!String methodsFor: 'converting' stamp: 'gzi 7/1/2003 10:24'!
asDecimal 
	"Answer the Integer created by interpreting the receiver as the string representation of a Decimal. The string may contain grouping commas a decimal point and a leading minus sign"
^Decimal fromString:self! !

